#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.defs.terminal import NetTerminalDef
from harkmw.HarkError import HARKError

from collections import OrderedDict


class NetworkDef:
    """ HARK network definition """

    INPUT = 'INPUT'
    OUTPUT = 'OUTPUT'
    CONDITION = 'CONDITION'

    def __init__(self, document):
        self.document = document
        self.name = None
        self.description = None
        self.type = None
        self.topic = None
        self.nodes = OrderedDict()
        self.links = []
        self.inputs = OrderedDict()
        self.outputs = OrderedDict()
        self.conditions = OrderedDict()

        self.process = None

    def loadXML(self, element):
        """ Load network definition from XML """

        # Fetch network name
        if element.getAttributeNode('name'):
            self.name = element.attributes['name'].value
        else:
            raise HARKError('No network name')

        # Fetch description
        if element.getAttributeNode('description'):
            self.description = element.attributes['description']
        else:
            self.description = ''

        # Fetch type
        typeName = ''
        if element.getAttributeNode('type'):
            typeName = element.attributes['type'].value

        if typeName == '' or typeName == 'subnet':
            self.type = 'subnet'
        elif typeName == 'iterator':
            self.type = 'iterator'
        elif typeName == 'threaded':
            self.type = 'threaded'
        else:
            raise HARKError("Unexpected type '%s'" % typeName)

        # Fetch nodes
        for child in element.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.nodeName == 'Node':
                from harkmw.defs.node import NodeDef
                newNode = NodeDef(
                    self, self.document.typedefs if self.document is not None else None)
                newNode.loadXML(child)
                self.nodes[newNode.name] = newNode

        # Fetch links
        for child in element.childNodes:
            if child.nodeType == child.ELEMENT_NODE and child.nodeName == 'Link':
                nodeFromName = child.attributes['from'].value
                nodeToName = child.attributes['to'].value
                inputName = child.attributes['input'].value
                outputName = child.attributes['output'].value

                # Ignoring node position

                nodeFrom = self.findNode(nodeFromName)
                nodeTo = self.findNode(nodeToName)
                if nodeFrom is None:
                    raise HARKError(
                        "No node with name '%s' found" % nodeFromName)

                elif nodeTo is None:
                    raise HARKError(
                        "No node with name '%s' found" % nodeToName)

                else:
                    pinFrom = nodeFrom.findOutputTerminal(outputName)
                    pinTo = nodeTo.findInputTerminal(inputName)

                    if pinFrom is None:
                        pinFrom = nodeFrom.addOutputTerminal(outputName)
                    if pinTo is None:
                        pinTo = nodeTo.addInputTerminal(inputName)

                    from harkmw.defs.link import Link
                    newLink = Link(pinFrom, pinTo)
                    self.links.append(newLink)

        # Fetch NetInput/NetOutput/NetCondition
        inputs = OrderedDict()
        outputs = OrderedDict()
        conditions = OrderedDict()
        for child in element.childNodes:
            if child.nodeType == child.ELEMENT_NODE:
                if (child.nodeName == 'NetInput' or child.nodeName ==
                        'NetOutput' or child.nodeName == 'NetCondition'):
                    externalName = child.attributes['name'].value if child.getAttributeNode(
                        'name') else None
                    nodeTerminalName = child.attributes['terminal'].value if child.getAttributeNode(
                        'terminal') else None
                    nodeName = child.attributes['node'].value if child.getAttributeNode(
                        'node') else None

                    node = self.findNode(nodeName)
                    if node is None:
                        raise HARKError("Unexpected node name '%s'" % nodeName)

                    elif child.nodeName == 'NetInput':
                        pinInput = node.findInputTerminal(nodeTerminalName)
                        if pinInput is None:
                            # Nertwork 内の Sub-Network 定義は、Sub-Network
                            # 自体の定義に先んじて現れる。(以下同様)
                            pass

                        netTerminal = NetTerminalDef(
                            node, nodeTerminalName, NetworkDef.INPUT)
                        inputs[externalName] = netTerminal

                    elif child.nodeName == 'NetOutput':
                        pinOutput = node.findOutputTerminal(nodeTerminalName)
                        if pinOutput is None:
                            # Nertwork 内の Sub-Network 定義は、Sub-Network
                            # 自体の定義に先んじて現れる。(以下同様)
                            pass

                        netTerminal = NetTerminalDef(
                            node, nodeTerminalName, NetworkDef.OUTPUT)
                        outputs[externalName] = netTerminal

                    else:  # 'NetCondition'
                        pinOutput = node.findOutputTerminal(nodeTerminalName)
                        if pinOutput is None:
                            # Nertwork 内の Sub-Network 定義は、Sub-Network
                            # 自体の定義に先んじて現れる。(以下同様)
                            pass

                        netTerminal = NetTerminalDef(
                            node, nodeTerminalName, NetworkDef.CONDITION)
                        conditions[externalName] = netTerminal

                elif child.nodeName == 'Note':
                    # Ignored
                    None

        self.inputs = inputs
        self.outputs = outputs
        self.conditions = conditions

    def insert_network_params(self, parameters):
        for node in self.nodes.values():
            node.insert_params(parameters)

    def findNode(self, name):
        return self.nodes[name] if name in self.nodes else None

    def get_process_name(self):
        if self.process is not None:
            return self.process
        else:
            return 'local'

    def dump(self, prefix=''):
        print(prefix + "[Network]")
        print(prefix + "  name=%s" % self.name)
        print(prefix + "  description=%s" % self.description)
        print(prefix + "  type=%s" % self.type)

        print(prefix + "  [Nodes]")
        for name, node in self.nodes.items():
            node.dump(prefix + "  ")

        print(prefix + "  [Links]")
        for link in self.links:
            link.dump(prefix + "  ")

        print(prefix + "  [Inputs]")
        for nt in self.inputs.values():
            nt.dump(prefix + "  ")

        print(prefix + "  [Outputs]")
        for nt in self.outputs.values():
            nt.dump(prefix + "  ")

        print(prefix + "  [Conditions]")
        for nt in self.conditions.values():
            nt.dump(prefix + "  ")

# end of file
