#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.HarkError import HARKError
from harkmw.defs.network import NetworkDef
from harkmw.defs.parameter import Parameters

import xml.dom.minidom
import os


class Input:
    def __init__(self):
        self.name = None
        self.type = None
        self.value = None
        self.description = None


class Output:
    def __init__(self):
        self.name = None
        self.type = None
        self.value = None
        self.description = None


class Typedef:
    def __init__(self):
        self.inputs = {}
        self.outputs = {}
        self.parameters = Parameters()


class Typedefs:
    def __init__(self):
        self.typedefs = {}
        self.nodedefs = {}

    def load_typedef(self, path):
        """ Load typelib definition from file .def file
        """

        # Load into memory
        defFileString = ''
        with open(path) as f:
            isFirstLine = True
            for line in f:
                if not isFirstLine or "#!" not in line:
                    defFileString += line
                isFirstLine = False

        # Parse XML
        dom = xml.dom.minidom.parseString(defFileString)
        self.load_typedef_xml(dom.documentElement)

    def load_nodedef(self, path):
        """ Load node definition from file .n file
        """

        # Load into memory
        nFileString = ''
        with open(path) as f:
            isFirstLine = True
            for line in f:
                if not isFirstLine or "#!" not in line:
                    nFileString += line
                isFirstLine = False

        # Parse XML
        dom = xml.dom.minidom.parseString(nFileString)

        for node in dom.documentElement.childNodes:
            # and node.attributes['name'] == 'MAIN' :
            if node.nodeName == 'Network':
                name = os.path.basename(path).split('.')[0]
                self.load_nodedef_xml(node, name)

    def load_typedef_xml(self, node):
        """ Load typelib definition from XML
        """
        # Process each 'depth=1' elements
        for child in node.childNodes:
            if child.nodeType != child.ELEMENT_NODE:
                continue

            if child.nodeName == 'NodeClass':
                typedef = Typedef()
                typedef.name = child.attributes['name'].value if child.getAttributeNode(
                    'name') else None
                typedef.category = child.attributes['category'].value if child.getAttributeNode(
                    'category') else 'Unknown'
                typedef.source = child.attributes['source'].value if child.getAttributeNode(
                    'source') else None
                typedef.require = child.attributes['require'].value if child.getAttributeNode(
                    'require') else None

                for cchild in child.childNodes:
                    if cchild.nodeType != cchild.ELEMENT_NODE:
                        continue

                    if cchild.nodeName == 'Input':
                        input = Input()
                        input.name = cchild.attributes['name'].value if cchild.getAttributeNode(
                            'name') else None
                        input.type = cchild.attributes['type'].value if cchild.getAttributeNode(
                            'type') else None
                        input.value = cchild.attributes['value'].value if cchild.getAttributeNode(
                            'value') else None
                        # ignore description value
                        typedef.inputs[input.name] = input

                    elif cchild.nodeName == 'Output':
                        output = Output()
                        output.name = cchild.attributes['name'].value if cchild.getAttributeNode(
                            'name') else None
                        output.type = cchild.attributes['type'].value if cchild.getAttributeNode(
                            'type') else None
                        output.value = cchild.attributes['value'].value if cchild.getAttributeNode(
                            'value') else None
                        # ignore description value
                        typedef.outputs[output.name] = output

                    elif cchild.nodeName == 'Parameter':
                        name = cchild.attributes['name'].value if cchild.getAttributeNode(
                            'name') else None
                        type = cchild.attributes['type'].value if cchild.getAttributeNode(
                            'type') else None
                        value = cchild.attributes['value'].value if cchild.getAttributeNode(
                            'value') else None
                        description = ''
                        # ignore description and list, valid attribute
                        typedef.parameters.add(name, type, value, description)

                    elif cchild.nodeName == 'Description':
                        # ignored
                        pass

                    else:
                        raise HARKError(
                            "Unexpected node <%s> in definition" %
                            child.nodeName)

                self.typedefs[typedef.name] = typedef

            elif child.nodeName == 'ModuleDepend':
                # ignored
                pass

            elif child.nodeName == 'FileDepend':
                # ignored
                pass

            else:
                raise HARKError(
                    "Unexpected node <%s> in definition" % child.nodeName)

    def load_nodedef_xml(self, node, name):
        """ Load node definition from XML
        """
        node_def = NetworkDef(None)
        node_def.loadXML(node)
        node_def.name = "subnet_%s" % name
        self.nodedefs[name] = node_def

    def dump(self):
        for name, value in self.typedefs.items():
            print("Type: %s" % name)

# end of file
