#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.logger import logger
import harkmw.logic.node
from collections import defaultdict
import os

debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class MQTTNode(harkmw.logic.node.Node):
    def __init__(self):
        harkmw.logic.node.Node.__init__(self)
        self.inputs = {}
        self.outputs = defaultdict(lambda: [])
        self.data = defaultdict(lambda: {})
        self.inputLookWindows = defaultdict(lambda: (0, 0))
        self.outputLookWindows = defaultdict(lambda: (0, 0))
        self.network = None

    def build(self, network, node_def, params):
        self.network = network
        self.type = node_def.type
        self.name = node_def.name
        self.topic = node_def.topic
        self.process_name = node_def.get_process_name()

    def process(self, output, count):
        pass

    def get_output(self, output_name, count):
        if output_name in self.data:
            output_data = self.data[output_name]
            if count in output_data:
                return output_data[count]
            else:
                raise BaseException(
                    'MQTTNode.get_output: no data with count {} in data with name {}'.format(
                        count, output_name))

        else:
            raise BaseException(
                'MQTTNode.get_output: no data with name {}, type={}, name={}'.format(
                    output_name, self.type, self.name))

    def get_input_count(self):
        return len(self.inputs)

    def get_output_count(self):
        return len(self.outputs)

    def get_input_look_window(self, input_id):
        return self.inputLookWindows[input_id]

    def set_input_look_window(self, input_id, lookWindow):
        self.inputLookWindows[input_id] = lookWindow

    def get_output_look_window(self, output_name):
        return self.outputLookWindows[output_name]

    def set_output_look_window(self, output_id, lookWindow):
        self.outputLookWindows[output_id] = lookWindow

    def connect_node(self, terminal_name, node_from, terminal_from_name):
        pass

    def push_data(self, input, count, process, internal=False, outer_count=-1):
        logger.debug(
            "MQTTNode.push_data: Try push data for Node [{}/{}], count [{}/{}], #inputs={}, #outputs={}, ({})".format(
                self.type, self.name, count, self.count, len(
                    self.inputs), len(
                    self.outputs), self))

        # if False:
        #     process.publish(self.get_topic() + "/_INPUT", count, True)
        #
        # else:
        #     process.subscribe(self.get_topic() + "/_INPUT/ack", self)
        #     process.publish(self.get_topic() + "/_INPUT", count, True)
        #
        #     # wait for response
        #     process.wait_topic(self.get_topic() + "/_INPUT/ack")
        #     process.unsubscribe(self.get_topic() + "/_INPUT/ack", self)

        if len(self.inputs) == 0:
            t = '{}/_INPUT'.format(self.get_topic())
            tack = '{}/ack'.format(t)
            terr = '{}/err'.format(t)
            process.subscribe([tack, terr], None)
            process.publish(t, count, True)

            # wait for response
            try:
                topic, data = process.wait_topic([tack, terr])
                if topic == tack:
                    pass
                elif topic == terr:
                    logger.error('Remote process returned error: ')
                    for d in data["data"]:
                        for i, l in enumerate(str(d).splitlines()):
                            logger.error(
                                '- {}{}'.format(("- " if i > 0 else ""), l))

                    raise Exception(
                        'Remote process returned error: {}'.format(
                            data["data"]))

            finally:
                process.unsubscribe([tack, terr], None)

        # logger.debug("MQTTNode.push_data: return: Node [{}/{}], count
        # [{}/{}], #inputs={}, #outputs={}, ({})".format(self.type, self.name,
        # count, self.count, len(self.inputs), len(self.outputs), self))
        return count

    def dump(self):
        print('[MQTTNode]')
        print('  type: %s' % self.type)
        print('  name: %s' % self.name)
        print('  network: type=%s, name=%s' %
              (self.network.type, self.network.name))

        print('  inputs: ')
        for (name, terminal) in self.inputs.items():
            print(
                '    Input: name=%s, node=%s/%s, terminal=%s' %
                (name,
                 terminal.input_node.type,
                 terminal.input_node.name,
                 terminal.input_terminal_name))

        print('  outputs: ')
        for (name, proxy_list) in self.outputs.items():
            for proxy in proxy_list:
                print(
                    '    Output: name=%s, node=%s/%s, terminal=%s' %
                    (name,
                     proxy.output_node.type,
                     proxy.output_node.name,
                     proxy.output_terminal_name))

    def __str__(self):
        return "<%s: type=%s, name=%s>" % (
            self.__class__.__name__, self.type, self.name)

# end of file
