#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import threading

from harkmw.HarkError import HARKError
from harkmw.logic.mqtt_node import MQTTNode
from harkmw.types.native_node import NativeNode

from harkmw.logger import logger

import harkmwnative

import os
debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class ProxyTerminal:
    def __init__(self, node, terminal):
        self.node = node
        self.terminal = terminal
        self.proxy = None

    def __str__(self):
        return "<ProxyTerminal: node=%s/%s, terminal=%s, proxy=%s>" % (
            self.node.type, self.node.name, self.terminal, self.proxy)


class Proxy:
    """ Base-class of Network, NativeProxy, Proxy
    """

    def __init__(self, process):
        self.native = harkmwnative.Proxy(self)
        self.process = process
        self.input_node = None
        self.output_node = None
        self.count = -1
        self.forced_count = -1

        self.buffer = []
        self.buffer_offset = 0
        self.buffer_lock = threading.Lock()

    def __str__(self):
        return "<Proxy: input=%s, output=%s>" % (
            self.input_node, self.output_node)

    def connect_native_input(self, node, terminal):
        """ この Proxy オブジェクトの入口側に、C++ 実装ノードの出口をつなぐ
        """
        if self.input_node is not None:
            raise HARKError("Proxy node already connected")

        self.native.connect_input(node.native_node, terminal)
        self.input_node = node
        self.input_terminal_name = terminal

        node.outputs[terminal].append(self)

    def connect_network_input(self, network, terminal):
        """ この Proxy オブジェクトの入口側に、ネットワークの出口をつなぐ
        """
        if self.input_node is not None:
            raise HARKError("Proxy node already connected")

        proxyTerminal = network.output_terminals[terminal]
        self.native.connect_input(
            proxyTerminal.node.native_node, proxyTerminal.terminal)
        self.input_node = proxyTerminal.node
        self.input_terminal_name = proxyTerminal.terminal

        network.output_collector[terminal].append(self)

    def connect_node_input(self, node, terminal):
        """ この Proxy オブジェクトの入口側に、C++ 実装でないノードの出口をつなぐ
        """
        if self.input_node is not None:
            raise HARKError("Proxy node already connected")

        self.input_node = node
        self.input_terminal_name = terminal

        node.outputs[terminal].append(self)

    def connect_native_output(self, node, terminal):
        """ この Proxy オブジェクトの出口側に、C++ 実装ノードの入口をつなぐ
        """
        if self.output_node is not None:
            raise HARKError("Proxy node already connected")

        self.native.connect_output(node.native_node, terminal)
        self.output_node = node
        self.output_terminal_name = terminal

    def connect_network_output(self, network, terminal):
        """ この Proxy オブジェクトの出口側に、ネットワークの入口をつなぐ
        """
        if self.output_node is not None:
            raise HARKError("Proxy node already connected")

        proxyTerminal = network.input_terminals[terminal]
        self.native.connect_output(
            proxyTerminal.node.native_node, proxyTerminal.terminal)
        self.output_node = proxyTerminal.node
        self.output_terminal_name = proxyTerminal.terminal

    def connect_node_output(self, node, terminal):
        """ この Proxy オブジェクトの出口側に、C++ 実装でないノードの入口をつなぐ
        """
        if self.output_node is not None:
            raise HARKError("Proxy node already connected")

        node.inputs[terminal] = self
        self.output_node = node
        self.output_terminal_name = terminal

    # def request(self, process):
    #     look_windows = [(0, 0)]
    #     # # if self.output_node is not None:
    #     # #     for (output_name, output) in self.output_node.outputs.items() :
    #     # #         logger.debug("output_name: {}, output={}, node={}".format(output_name, output, self.output_node))
    #     # #         look_windows.append(self.output_node.get_output_look_window(output_name))
    #     # look_window = max(look_windows)
    #     # look_window = [
    #     #     sum(x) for x in zip(
    #     #         look_window,
    #     #         self.output_node.get_input_look_window(
    #     #             self.output_terminal_name))]
    #     # logger.debug('Proxy.request: look_window: in={}'.format(list(zip(
    #     #     [(tname, self.output_node.get_input_look_window(tname)) for i in self.output_node.inputs.keys()]))))
    #     look_window = [max(x) for x in zip([self.output_node.get_input_look_window(
    #         tname) for i in self.output_node.inputs.keys()])]
    #     logger.debug('Proxy.request: look_window={}'.format(look_window))
    #
    #     logger.debug(
    #         "Proxy: request: self={}, look_window={}, look_windows={}".format(
    #             self, look_window, look_windows))
    #     self.set_look_window(look_window)
    #     if self.input_node is not None:
    #         self.input_node.request(process)

    def initialize(self, process):
        self.native.initialize(
            tag='Proxy/{}/{}'.format(self.input_node.name, self.output_node.name))

        logger.debug(
            'Initialize: proxy: input_node: {} / {} / {}, output_node: {} / {} / {}'.format(
                (self.input_node.name,
                 self.input_node.type),
                type(
                    self.input_node),
                self.input_node.process_name,
                (self.output_node.name,
                 self.output_node.type),
                type(
                    self.output_node),
                self.output_node.process_name))

        logger.debug(
            '{}, {}'.format(
                (self.input_node.process_name != process.process_name),
                (self.output_node.process_name == process.process_name)))

        if self.input_node.process_name != process.process_name and self.output_node.process_name == process.process_name:
            if self.output_node.process_name == process.process_name or \
                (self.output_node.name[-10:] == '_condition') or \
                (self.output_node.name[-7:] == '_output' and self.output_node.network.network_type == 'iterator') or \
                    (self.input_node.name[-6:] == '_input'):

                logger.debug('{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}'.format(self.output_node.process_name == process.process_name,
                                                                                 (self.output_node.name[-10:] == '_condition'),
                                                                                 self.output_node.name[-7:] == '_output',
                                                                                 self.output_node.network.type == 'iterator',
                                                                                 self.input_node.name[-6:] == '_input',
                                                                                 self.input_node.process_name,
                                                                                 self.output_node.process_name,
                                                                                 process.process_name,
                                                                                 self.input_node.name,
                                                                                 self.output_node.name,
                                                                                 self.output_node.network.network_type))
                """ Proxy の入口のノードはリモート実行、あるいは CONDITION 出力 """
                topic = "%s/%s" % (self.input_node.get_topic(),
                                   self.input_terminal_name)

                self.process.subscribe(topic + '/_OUTPUT', self)
                self.input_node.network.mqtt_proxies.append(self)

            if self.output_node.name[-6:] == '_input':
                topic = "%s/%s" % (self.output_node.get_topic(),
                                   self.output_terminal_name)
                self.process.subscribe(topic + '/_INPUT', self)

    def has_remote_data_source(self):
        if isinstance(self.input_node, MQTTNode):
            return True
        elif self.input_node.type == 'HARKMW_Collector':
            if self.input_terminal_name in self.input_node.inputs:
                return self.input_node.inputs[self.input_terminal_name].has_remote_data_source(
                )
        return False

    def push_data(self, count, process, body=None):
        logger.debug(
            'Proxy.push_data: output_node={}, count={}, body={}'.format(
                type(
                    self.output_node),
                count,
                (body is not None)))

        # if (isinstance(self.output_node, MQTTNode) or
        # isinstance(self.input_node, MQTTNode)) and body is not None:
        if ((isinstance(self.output_node, MQTTNode)
             or self.has_remote_data_source()) and (body is not None)):
            logger.debug(
                'Proxy.push_data: native.set_data: count={}'.format(count))
            try:
                self.native.set_data(count, body)
            except Exception as ex:
                logger.error(ex)
            logger.debug('Proxy.push_data: native.set_data: end')

        if isinstance(self.output_node, MQTTNode):
            """ Proxy の出口のノードはリモート実行 """
            logger.debug(
                'Proxy.push_data: output-node is remote: native.get_data: count={}'.format(count))
            data = self.native.get_data(count)
            logger.debug(
                'Proxy.push_data: native.get_data: data={}'.format(data))

            count = self.output_node.push_data(self, count, process)

        else:
            logger.debug(
                'Proxy.push_data: output-node is native: native.push_data: count={}'.format(count))
            count = self.native.push_data(count)
            logger.debug('Proxy.push_data: native.push_data: end')

            self.count = count

            if self.output_node.type == "HARKMW_Collector" and self.output_node.name[-6:] == "_input":
                logger.debug('Proxy.push_data: push to next network')
                self.output_node.network.push_data(self, count, process)
            else:
                logger.debug('Proxy.push_data: push to next node')
                count = self.output_node.push_data(
                    self, count, process)
                # look_window = self.get_look_window()
                # if count >= look_window[0]:
                #     count = self.output_node.push_data(
                #         self, count - look_window[0], process)
                # else:
                #     count = -1
        return count

    def get_data(self, count):
        if isinstance(self.input_node, NativeNode):
            return self.input_node.get_data(count)
        else:
            return self.input_node.get_output(self.input_terminal_name, count)

    def get_input_node(self):
        return self.input_node

    def get_output_node(self):
        return self.output_node

    def get_input_terminal_name(self):
        return self.input_terminal_name

    def get_output_terminal_name(self):
        return self.output_terminal_name

    def force_count(self, count):
        self.forced_count = count
        self.native.force_count(count)

    def get_look_window(self):
        return self.native.get_look_window()

    def set_look_window(self, look_window):
        return self.native.set_look_window(look_window[0], look_window[1])

    def get_delay(self):
        if self.input_node is None:
            """ 他のプロセスでノードが動いている場合
            """
            return 0

        elif self.input_node.type == "Collector" and self.input_node.name[-6:] == "_input":
            return self.native.get_look_window()[0]

        else:
            return self.input_node.get_delay(
            ) + self.native.get_look_window()[0]

# end of file
