#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import sys
import os
import pkg_resources
import argparse

import harkmw
from harkmw.logger import initialize_log, logger


def version(dirs):
    app = harkmw.Process()
    app.print_modules(dirs)


default_toolbox_dir_nt = 'C:/Program Files/HARK/lib/hark/toolbox'
default_toolbox_dir_default = '/usr/local/lib/hark/toolbox:/usr/lib/hark/toolbox'

default_toolbox_dir = os.getenv('HARK_PATH') if os.getenv(
    'HARK_PATH') else default_toolbox_dir_nt if os.name == 'nt' else default_toolbox_dir_default

default_process_name = 'local'
default_topic = 'harkmw'

default_daemon_name = 'harmkw-daemon'
default_daemon_pidfile = '/var/run/harkmwd.pid'


def app_main():
    parser = argparse.ArgumentParser(description='HARK middleware')

    parser.add_argument(
        '--toolbox',
        '-t',
        metavar='DIR',
        type=str,
        help='Load toolbox from DIR.  If omitted, default value %(default)s is used.',
        default=default_toolbox_dir)

    parser.add_argument(
        '--process',
        '-p',
        metavar='NAME',
        type=str,
        help='Run as NAME.  If omitted, default value \'%(default)s\' is used.',
        default=default_process_name)

    parser.add_argument('--broker', '-b', type=str,
                        help='MQTT Broker host:port.', default='localhost')

    parser.add_argument('--topic', '-c', type=str,
                        help='MQTT topic base.', default=default_topic)

    parser.add_argument(
        '--local',
        '-l',
        action='store_true',
        help='Run on local host.  Treat all "process" value as "local".',
        default=False)

    parser.add_argument(
        '--startup',
        '-s',
        action='store_true',
        default=False,
        help='Before start process, request harkmw-daemon to startup processes.')

    parser.add_argument(
        '--log-config',
        type=str,
        help='Configuration file for python logging mechanism',
        default=None)

    parser.add_argument(
        '--log-level',
        type=str,
        help='Log level, ignored if --log-config is given.',
        default='info')

    parser.add_argument('--output-file', type=str, metavar='FILE',
                        help='Output log to FILE.', default=None)

    parser.add_argument(
        '--output-fd',
        type=int,
        metavar='NUM',
        help='Output log to file descriptor NUM.',
        default=None)

    parser.add_argument(
        '--version',
        '-v',
        action='store_true',
        help='Show module version information and exit.',
        default=False)

    parser.add_argument(
        '--mwversion',
        '-V',
        action='version',
        help='Show hark-middleware version information and exit.',
        version='HARK middleware {}'.format(
            pkg_resources.get_distribution('harkmw').version))

    parser.add_argument('network', nargs='?',
                        help='Load .n file from NETWORK.')

    parser.add_argument('arg', nargs='*',
                        help='Arguments to be passed to network.')

    args = parser.parse_args()
    initialize_log(args.log_config, args.log_level,
                   output_file=args.output_file, output_fd=args.output_fd)

    # Set option values
    network_file = args.network

    if os.name == 'nt':
        toolbox_dirs = args.toolbox.split(";")
    else:
        toolbox_dirs = args.toolbox.split(":")

    if args.version:
        version(toolbox_dirs)
        sys.exit(0)

    process_name = args.process

    if network_file is None:
        parser.print_help(sys.stderr)
        sys.exit(1)

    # Dump options
    logger.info("Start HARK middleware:")
    logger.info("    network: {}".format(network_file))
    logger.info("    process: {}".format(process_name))
    logger.info("    toolbox: {}".format(toolbox_dirs))
    logger.info("    args:    {}".format(args.arg))

    # Create application instance
    app = harkmw.Process(args.arg, broker=args.broker, topic=args.topic)
    app.add_toolbox_dirs(toolbox_dirs)
    app.load(network_file)
    app.run(process_name, local=args.local, startup=args.startup)


def daemon_main():
    # requires python-daemon.  (try: $ pip install python daemon)
    import daemon
    # requires python-daemon.  (try: $ pip install python daemon)
    import daemon.pidfile

    parser = argparse.ArgumentParser(description='HARK middleware daemon')
    parser.add_argument(metavar='NAME', nargs='*', dest='process_names',
                        help='List of process name.')

    parser.add_argument('--broker', '-b', type=str,
                        help='MQTT Broker host:port.', default='localhost')

    parser.add_argument('--topic', '-c', type=str,
                        help='MQTT topic base.', default=default_topic)

    parser.add_argument(
        '--toolbox',
        '-t',
        metavar='DIR',
        type=str,
        help='Load toolbox from DIR.  If omitted, default value %(default)s is used.',
        default=default_toolbox_dir)

    parser.add_argument('--workdir', '-d', type=str,
                        help='Working directory.', default='/var/lib/harkmw')

    parser.add_argument(
        '--log-config',
        type=str,
        help='Configuration file for python logging mechanism',
        default=None)

    parser.add_argument(
        '--log-level',
        type=str,
        help='Log level, ignored if --log-config is given.',
        default='info')

    parser.add_argument('--output-file', type=str, metavar='FILE',
                        help='Output log to FILE.', default=None)

    parser.add_argument(
        '--output-fd',
        type=int,
        metavar='NUM',
        help='Output log to file descriptor NUM.',
        default=None)

    parser.add_argument('--no-daemon', '-n', action='store_true',
                        help='Do not work as daemon process.')

    parser.add_argument('--pid-file', type=str, default=default_daemon_pidfile,
                        help='PID file path.')

    parser.add_argument(
        '--version',
        '-v',
        action='store_true',
        help='Show module version information and exit.',
        default=False)

    parser.add_argument(
        '--mwversion',
        '-V',
        action='version',
        help='Show hark-middleware version information and exit.',
        version='HARK middleware {}'.format(
            pkg_resources.get_distribution('harkmw').version))

    args = parser.parse_args()
    if args.no_daemon:
        pid = None
    else:
        pid = daemon.pidfile.PIDLockFile(args.pid_file)
        if pid.is_locked():
            raise Exception('Process is already started.')
        # stdout = open('out', 'w')

    if os.name == 'nt':
        toolbox_dirs = args.toolbox.split(";")
    else:
        toolbox_dirs = args.toolbox.split(":")

    dp = harkmw.Daemon(broker=args.broker, topic=args.topic)
    dp.add_toolbox_dirs(toolbox_dirs)
    if args.no_daemon:
        initialize_log(args.log_config, args.log_level,
                       output_file=args.output_file, output_fd=args.output_fd)

        dp.run(args.process_names)
    else:
        with daemon.DaemonContext(working_directory=args.workdir, pidfile=pid):
            initialize_log(args.log_config, args.log_level,
                           output_file=args.output_file, output_fd=args.output_fd)

            dp.run(args.process_names)


if __name__ == '__main__':
    app_main()

# end of file
