#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

import pickle
import sys

from harkmw.defs.document import Document
from harkmw.defs.type import Typedefs
from harkmw.defs.parameter import Parameters
from harkmw.logic.builder import Builder
from harkmw.logic.network import Network
from harkmw.logic.proxy import Proxy
from harkmw.logic.node import Node
from harkmw.types.dynamic import Dynamic

from harkmw.logger import logger
from harkmw.HarkError import HARKError

import os
debug = (bool(os.environ['HARK_DEBUG'])
         ) if 'HARK_DEBUG' in os.environ else False


class Process:
    def __init__(self, args=[], broker='localhost', topic='harkmw'):
        self.typedefs = Typedefs()

        params = Parameters()
        for i, argv in enumerate(args):
            params.add('ARG%d' % (i + 1), 'string', argv)
            params.add('string:ARG%d' % (i + 1), 'string', argv)
            # ToDo (in C++, = Int::alloc(atoi(argv[arg])))
            params.add('int:ARG%d' % (i + 1), 'int', argv)
            # ToDo (in C++, = Float::alloc(atof(argv[arg])))
            params.add('float:ARG%d' % (i + 1), 'float', argv)
            # ToDo (in C++, see batchflow.cc:75)
            params.add('object:ARG%d' % (i + 1), 'object', argv)
#       params.dump()
        self.args = args
        self.broker = broker
        self.topic = topic

        self.params = params
        self.toolbox_dirs = []
        self.mosquitto_client = None
        self.mqtt_topics = {}
        self.mqtt_queue = []

    def add_toolbox_dirs(self, dirs):
        self.toolbox_dirs.extend(dirs)

    def parse(self, ndata):
        # Scan toolboxes
        for dir in self.toolbox_dirs:
            self.scan_nodedefs(dir)
            self.scan_typedefs(dir)
            self.scan_typelibs(dir)

        # Load .n/.x file
        document = Document(self.typedefs)
        document.parse(ndata)
        self.ndata = ndata
        self.document = document

    def load(self, nfile):
        self.nfile = nfile

        # Load into memory
        nfileString = ''
        with open(nfile) as f:
            isFirstLine = True
            for line in f:
                if not isFirstLine or "#!" not in line:
                    nfileString += line
                isFirstLine = False

        self.parse(nfileString)

    def run(self, process_name, local=False, startup=None, notify_init=None):
        logger.debug(
            'Process.run: process_name={}, local={}, startup={}, notify_init={}'.format(
                process_name, local, startup, notify_init))
        self.process_name = process_name
        self.to_quit = False
        logger.debug('set process.to_quit: False')

        # Build network
        builder = Builder(self)
        network = builder.build(
            self.document, self.params, process_name, force_local=local)

        try:
            self.running = True
            network.request(self)
            network.initialize(self)
            if network.has_remote_nodes(self):
                self.subscribe('_TERMINATE', self)

            if startup:
                self.startup_processes(network)
            if notify_init:
                self.publish(notify_init, 0, {})

            # Run network
            network.run(self)
            network.release()

        except Exception as ex:
            import traceback
            traceback.print_exc()
            sys.exit(1)

        finally:
            self.running = False
            if self.is_mqtt_initialized():
                self.finalize_mqtt()

    def startup_processes(self, network):
        # scan remote process names
        def scan_network_processes(network):
            ps = set()

            for node in network.nodes.values():
                if isinstance(node, Network):
                    ps2 = scan_network_processes(node)
                    ps.update(ps2)
                else:
                    ps.add(node.process_name)

            return ps

        processes = list(scan_network_processes(network))

        payload = {
            'ndata': self.ndata,
            'args': self.args,
        }
        payload = pickle.dumps(payload)

        for p in processes:
            if p == 'local':
                continue

            req_topic = '{}/_DAEMON/req/start/{}'.format(self.topic, p)
            res_topic = '{}/res/start'.format(p)
            self.subscribe('{}/status'.format(res_topic), self)
            self.mosquitto_client.publish(req_topic, payload, 0)

            # wait for response
            self.wait_topic('{}/status'.format(res_topic))
            self.unsubscribe('{}/status'.format(res_topic), self)

    def initialize_mqtt(self):
        # Setup MQTT
        import paho.mqtt.client as mqtt
        self.mosquitto_subscribers = {}
        self.mosquitto_client = mqtt.Client()
        self.mosquitto_client.user_data_set(self)
        self.mosquitto_client.on_message = self.on_mqtt_message

        if True:
            t = '{}/_TERMINATE'.format(self.topic)
            p = self.make_payload(t, 0, True)
            self.mosquitto_client.will_set(t, p, 0)

        args = {}
        if ':' in self.broker:
            host, port = self.broker.split(':')
            args['port'] = int(port)
        else:
            host = self.broker

        # self.document.mqtt_host, self.document.mqtt_port)
        self.mosquitto_client.connect(host, **args)

    def finalize_mqtt(self):
        # Close MQTT
        while self.mosquitto_client.want_write():
            self.mosquitto_client.loop_write()
        self.mosquitto_client.loop()
        self.mosquitto_client.disconnect()
        self.mosquitto_client = None

    def is_mqtt_initialized(self):
        return self.mosquitto_client is not None

    def on_mqtt_message(self, mosquitto, obj, msg):
        logger.debug('Process.on_mqtt_message: topic={}'.format(msg.topic))
        self.mqtt_queue.append(msg)

    def process_mqtt_message(self, msg):
        payload = msg.payload
        # marker = payload[0:4]
        # version = payload[4:8]
        # type = payload[8:12]
        data = payload[12:]

        obj = pickle.loads(data)
        if not isinstance(obj, dict):
            raise HARKError("Received non-dictionary data via MQTT")

        topic = obj['topic']
        count = obj['count']
        body = obj['data']

        logger.debug(
            'Receive: topic={:s}, count={:d}, msg.topic={}'.format(
                topic, count, msg.topic))

        if topic in self.mosquitto_subscribers:
            error = []
            for receiver, args in self.mosquitto_subscribers[topic]:
                self.mqtt_topics[topic] = obj
                if topic == '{}/_TERMINATE'.format(self.topic):
                    self.to_quit = True
                    logger.debug('set process.to_quit: True')

                elif isinstance(receiver, Proxy):
                    try:
                        logger.debug(
                            'Receiver is proxy: {}, topic={}'.format(
                                receiver, topic))
                        receiver.push_data(count, self, body)
                        logger.debug(
                            'Returned from proxy: success, receiver={}, topic={}'.format(
                                receiver, topic))
                    except Exception as ex:
                        error.append(str(ex))
                        logger.error(
                            'Error during proxy process: {}, receiver={}'.format(
                                ex, receiver))

                elif isinstance(receiver, Network):
                    try:
                        logger.debug(
                            'Receiver is network: {}'.format(receiver))
                        receiver.proceed(count, self)
                        logger.debug(
                            'Returned from network: success, receiver={}'.format(receiver))
                    except Exception as ex:
                        error.append(str(ex))
                        logger.error(
                            'Error during network process: {}, receiver={}'.format(
                                ex, receiver))

                elif isinstance(receiver, Node):
                    try:
                        logger.debug('Receiver is node: {}'.format(receiver))
                        receiver.push_data('', count, self)
                        logger.debug(
                            'Returned from node: success, receiver={}'.format(receiver))
                    except Exception as ex:
                        error.append(str(ex))
                        logger.error(
                            'Error during node process: {}, receiver={}'.format(
                                ex, receiver))

            if topic[-4:] != '/ack':
                if len(error) == 0:
                    self.publish(msg.topic[len(self.topic) + 1:] +
                                 '/ack', count, None)
                else:
                    self.publish(msg.topic[len(self.topic) + 1:] +
                                 '/err', count, error)

        else:
            logger.info('Topic "{}" is not in subscriber list.'.format(topic))

    def check_mqtt(self):
        if self.running:
            self.mosquitto_client.loop(.1)
            while len(self.mqtt_queue) > 0:
                self.process_mqtt_message(self.mqtt_queue.pop())

    def wait_topic(self, topic):
        torg = {}
        for i, t in enumerate(topic if isinstance(topic, list) else [topic]):
            tt = '{}/{}'.format(self.topic, t)
            torg[tt] = t
        topics = list(torg.keys())

        logger.debug('Waiting for topic: {}'.format(topics))
        while self.running and not self.to_quit:
            i = list(set(topics) & set(self.mqtt_topics.keys()))
            if i:
                tt = i[0]
                logger.debug('Found topic: {}'.format(tt))
                return torg[tt], self.mqtt_topics.pop(tt)

            else:
                self.check_mqtt()

    def make_payload(self, topic, count, data):
        return bytes('HARK' + '0001' + 'PCKL', encoding='utf8') + \
            pickle.dumps({'count': count, 'data': data, 'topic': topic})

    def publish(self, topic, count, data):
        if self.running:
            t = '{}/{}'.format(self.topic, topic)
            logger.debug("Publish: %s, %d" % (t, count))
            p = self.make_payload(t, count, data)
            result, qid = self.mosquitto_client.publish(t, p, 0)

    def subscribe(self, topic, receiver, args=None):
        if self.running:
            if not isinstance(topic, list):
                topics = [topic]
            else:
                topics = topic

            if not self.is_mqtt_initialized():
                self.initialize_mqtt()

            for topic in topics:
                t = '{}/{}'.format(self.topic, topic)

                if t not in self.mosquitto_subscribers:
                    self.mosquitto_subscribers[t] = []

                self.mosquitto_subscribers[t].append((receiver, args))

                self.mosquitto_client.subscribe(t, 0)

                if isinstance(receiver, Proxy):
                    logger.debug(
                        "Subscribe: %s -> %s/%s" %
                        (t, receiver.output_node.name, receiver.output_terminal_name))
                else:
                    logger.debug("Subscribe: %s -> %s" % (t, receiver))

    def unsubscribe(self, topic, receiver):
        if self.running:
            if not isinstance(topic, list):
                topics = [topic]
            else:
                topics = topic

            if not self.is_mqtt_initialized():
                self.initialize_mqtt()

            for topic in topics:
                if topic not in self.mosquitto_subscribers:
                    self.mosquitto_subscribers[topic] = []

                for i, r in enumerate(self.mosquitto_subscribers[topic]):
                    if r == receiver:
                        del self.mosquitto_subscribers[topic][i]
                if len(self.mosquitto_subscribers[topic]) == 0:
                    self.mosquitto_client.unsubscribe(topic)
                    del self.mosquitto_subscribers[topic]

                if isinstance(receiver, Proxy):
                    logger.debug(
                        "Unsubscribe: %s -> %s/%s" %
                        (topic, receiver.output_node.name, receiver.output_terminal_name))
                else:
                    logger.debug("Unsubscribe: %s -> %s" % (topic, receiver))

    def scan_nodedefs(self, path):
        import os
        for (root, dirs, files) in os.walk(path, True, None, True):
            for file in files:
                if file[-2:] == ".n":
                    self.typedefs.load_nodedef(os.path.join(root, file))

    def scan_typelibs(self, path):
        import os
        for (root, dirs, files) in os.walk(path, True, None, True):
            for file in files:
                if "tlb" in file.split("."):
                    Dynamic().load_dynamic(os.path.join(root, file))

    def scan_typedefs(self, path):
        import os
        for (root, dirs, files) in os.walk(path, True, None, True):
            for file in files:
                if file[-4:] == ".def":
                    self.typedefs.load_typedef(os.path.join(root, file))

    def print_modules(self, dirs):
        for path in dirs:
            self.scan_typelibs(path)
        Dynamic().print_modules()

# end of file
