/* -*- mode: C++; coding: utf-8 -*- */
/*
 *
 *  HARK Middleware
 *  Copyright (C) 2013-2014 by Honda Research Institute Japan Co., Ltd.
 *
 *  $Id:$
 */

#include "libharkmw.h"
#include "node.h"

#include <hark-base/object_param.h>

#include "type.h"
#include "collector.h"

namespace HARKMW
{
    static PyObject *
    Node_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
    {
        Node *self;

        self = (Node *)type->tp_alloc(type, 0);
        if(self != NULL) {
            self->node = NULL;
        }

        return (PyObject *)self;
    }

    static void
    Node_dealloc(Node* self)
    {
        if(self->node){
            delete self->node;
        }

        Py_TYPE(self)->tp_free((PyObject*)self);
    }

    static int
    Node_init(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);
        PyObject *nodeObj = NULL, *nameObj = NULL, *parametersObj = NULL;

        /* Parse args */
        static char *kwlist[] = {"node", "name", "parameters", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "OOO", kwlist, &nodeObj, &nameObj, &parametersObj)){
            fprintf(stdout, "Failed to parse args\n");
            return -1;
        }

        const char *name = NULL;
        if(nameObj){
            name = PyStr_AsUTF8(nameObj);
        }

        HARK::_NodeFactory *factory = HARK::Node::getFactoryNamed(name);
        if(!factory){
            std::stringstream ss;
            ss << "No node class with name '" << name << "' found";
            PyErr_SetString(PyExc_NotImplementedError, ((const std::string &)ss).c_str());
            return -1;
        }

        PyObject *paramsObj = PyObject_HasAttrString(parametersObj, "params") ? PyObject_GetAttrString(parametersObj, "params") : NULL;

        HARK::ParameterSet parameters;
        if(!paramsObj){
            fprintf(stdout, "arg 'parameters' does not have attribute named 'params'\n");

        }else{
            if(!PyDict_Check(paramsObj)){
                fprintf(stdout, "'parameters' is not a dictinary\n");
                return -1;
            }
            PyObject *parameterList = PyDict_Items(paramsObj);
            if(!PyList_Check(parameterList)){
                fprintf(stdout, "'parameters.items()' is not a list\n");
                return -1;
            }

            Py_ssize_t parameterCount = PyList_Size(parameterList);

            for(int i = 0; i < parameterCount; i++){
                PyObject *parameterTuple = PyList_GetItem(parameterList, i);
                if(!PyTuple_Check(parameterTuple)){
                    fprintf(stdout, "'parameters.items()[]' is not a tuple\n");
                    return -1;
                }

                PyObject *parameterName  = PyTuple_GetItem(parameterTuple, 0);
                PyObject *parameterParam = PyTuple_GetItem(parameterTuple, 1);
                if(!PyUnicode_Check(parameterName)){
                    fprintf(stdout, "'parameters.items()[][0]' is not a unicode (may be %s)\n", Py_TYPE(parameterName)->tp_name);
                    return -1;
                }

                PyObject *parameterType  = PyObject_HasAttrString(parameterParam, "type") ? PyObject_GetAttrString(parameterParam, "type") : NULL;
                PyObject *parameterValue = PyObject_HasAttrString(parameterParam, "value") ? PyObject_GetAttrString(parameterParam, "value") : NULL;
                if(!parameterType){
                    fprintf(stdout, "parameter.items()[][1][type] not found.\n");
                    return -1;
                }
                if(!parameterValue){
                    fprintf(stdout, "parameter.items()[][1][value] not found.\n");
                    return -1;
                }
                if(!PyUnicode_Check(parameterType) && !PyStr_Check(parameterType)){
                    fprintf(stdout, "'parameters.items()[][1][type]' is not a unicode/string (type=%s)\n", Py_TYPE(parameterType)->tp_name);
                    return -1;
                }
                if(!PyUnicode_Check(parameterValue) && !PyStr_Check(parameterValue)){
                    fprintf(stdout, "'parameters.items()[][1][value]' is not a unicode/string (type=%s)\n", Py_TYPE(parameterValue)->tp_name);
                    return -1;
                }

                if(strlen(PyStr_AsUTF8(parameterValue)) == 0){
                    continue;
                }

                parameters.add(PyStr_AsUTF8(parameterName),
                               HARK::ObjectParam::stringParam(PyStr_AsUTF8(parameterType), PyStr_AsUTF8(parameterValue), parameters));
            }
        }

        /* Do processes */
        try{
	    try{
	        self->node = factory->Create(name, parameters);
	    }catch(HARK::BaseException *ex){
	        std::stringstream ss;
		ss << "Exception while creating " << name;
		throw new HARK::GeneralException(ss.str(), __FILE__, __LINE__);
	    }
	}HARKMW_CATCH_EXCEPTION_RETURNS(-1);

        Py_INCREF(nodeObj);
        self->nodeObject = nodeObj;

        return 0;
    }

    static PyObject *
    Node_process(PyObject* _self, PyObject *args, PyObject *kwds)
    {
      // std::cout << "Node_process" << std::endl;
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        PyObject *outputNameObj = NULL;
        int count = 0;
        int input_count = -1;
        int output_count = -1;

        static char *kwlist[] = {"output_name", "count", "input_count", "output_count", NULL};
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "Oi|ii", kwlist, &outputNameObj, &count, &input_count, &output_count)){
            Py_RETURN_NONE;
        }

        if(!outputNameObj){
            fprintf(stdout, "Node_process: No output_name value given\n");
            Py_RETURN_NONE;
        }

        // fprintf(stdout, "Node_process: input-count=%d, output-count=%d\n", input_count, output_count);
        // fflush(stdout);

        const char *outputName = NULL;
        if(outputNameObj){
            outputName = PyStr_AsUTF8(outputNameObj);
        }

        HARK::ObjectRef output;
        try{
	    if(self->node->getName() == "HARKMW_Collector"){
		// fprintf(stdout, "collector\n");
		output = dynamic_cast<HARK::HARKMW_Collector *>(self->node)->getOutputNamed(outputName, input_count, output_count);
	    }else{
		// fprintf(stdout, "non-collector\n");
		output = self->node->getOutputNamed(outputName, count);
	    }

	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyObject *
    Node_get_output(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"name", "count", NULL};
        PyObject *argNameObj = NULL;
        int count;
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "Oi", kwlist, &argNameObj, &count)){
            // std::cout << "node.cc: Node_get_output: failed to parse args" << std::endl;
            Py_RETURN_NONE;
        }

        const char *name = NULL;
        if(argNameObj){
            name = PyStr_AsUTF8(argNameObj);
        }

        bool value = false;
        try{
            //int output_id = self->node->getOutputID(name);
            HARK::ObjectRef o = self->node->getOutputNamed(name, count);
//            std::cout << "node.cc: Node_get_output: name=" << name << ", count=" << count << ", result=" << o << std::endl;

            return toPythonObject(o);

        }HARKMW_CATCH_EXCEPTION();

        if(value){
            Py_RETURN_TRUE;
        }else{
            Py_RETURN_FALSE;
        }
    }

    static PyObject *
    Node_connect(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"in", "inputNode", "out", NULL};
        PyObject *argInObj = NULL, *argInputNodeObj = NULL, *argOutObj = NULL;
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "OOO", kwlist, &argInObj, &argInputNodeObj, &argOutObj)){
            fprintf(stdout, "Node_connect: Failed to parse args/keywords\n");
            Py_RETURN_NONE;
        }

        const char *inName = NULL, *outName = NULL;
        if(argInObj){
            inName = PyStr_AsUTF8(argInObj);
        }else{
            fprintf(stdout, "Node_connect: Arg 'in' is not a string\n");
            Py_RETURN_NONE;
        }
        if(argOutObj){
            outName = PyStr_AsUTF8(argOutObj);
        }else{
            fprintf(stdout, "Node_connect: Arg 'out' is not a string\n");
            Py_RETURN_NONE;
        }

        if(argInputNodeObj){
            if(!PyObject_TypeCheck(argInputNodeObj, &NodeType)){
                fprintf(stdout, "Node_connect: Arg 'inputNode' is not a Node object\n");
                Py_RETURN_NONE;

            }else{
                Node *inputNode = reinterpret_cast<Node *>(argInputNodeObj);

                try {
                    self->node->connectToNode(std::string(inName), inputNode->node, std::string(outName));

		}HARKMW_CATCH_EXCEPTION();
            }

        }else{
            fprintf(stdout, "Node_connect: Arg 'inputNode' is not given\n");
            Py_RETURN_NONE;
        }

        Py_RETURN_NONE;
    }

    static PyObject *
    Node_initialize(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        try {
            if(!self->node){
                fprintf(stdout, "Warn: node is NULL\n");
            }
            self->node->initialize();

	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyObject *
    Node_get_input_count(PyObject *_self, PyObject *args)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        return Py_BuildValue("i", self->node->inputs.size());
    }

    static PyObject *
    Node_get_output_count(PyObject *_self, PyObject *args)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        return Py_BuildValue("i", self->node->outputNames.size());
    }

    static PyObject *
    Node_get_output_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"output_id", NULL};
        PyObject *output_name_obj = NULL;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist, &output_name_obj)){
            Py_RETURN_NONE;
        }
        const char *output_name = NULL;
        if(output_name_obj){
            output_name = PyStr_AsUTF8(output_name_obj);
        }

	HARK::BufferedNode *bufferedNode = dynamic_cast<HARK::BufferedNode *>(self->node);
	int look_ahead = 0, look_back = 0;
	if(bufferedNode){
	    int output_id = bufferedNode->getOutputID(output_name);
	    //std::cout << "Node.cc: get_output_look_window: name=" << bufferedNode->getName() << ", outputName=" << output_name << ", output_id=" << output_id << ", lookWindow=(" << bufferedNode->outputs[output_id].lookAhead << ", " << bufferedNode->outputs[output_id].lookBack << ")" << std::endl;
	    look_ahead = bufferedNode->outputs[output_id].lookAhead;
	    look_back  = bufferedNode->outputs[output_id].lookBack;
	}

	return Py_BuildValue("(ii)", look_ahead, look_back);
    }

    static PyObject *
    Node_set_output_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"output_id", "lookAhead", "lookBack", NULL};
        PyObject *output_name_obj = NULL;
        int look_ahead, look_back;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "Oii", kwlist, &output_name_obj, &look_ahead, &look_back)){
            Py_RETURN_NONE;
        }
        const char *output_name = NULL;
        if(output_name_obj){
            output_name = PyStr_AsUTF8(output_name_obj);
        }

	HARK::BufferedNode *bufferedNode = dynamic_cast<HARK::BufferedNode *>(self->node);
	if(bufferedNode){
	    int output_id = bufferedNode->getOutputID(output_name);
	    //std::cout << "Node.cc: set_output_look_window: name=" << bufferedNode->getName() << ", outputName=" << output_name << ", output_id=" << output_id << ", lookWindow=(" << bufferedNode->outputs[output_id].lookAhead << ", " << bufferedNode->outputs[output_id].lookBack << ")" << std::endl;
	    bufferedNode->outputs[output_id].lookAhead = look_ahead;
	    bufferedNode->outputs[output_id].lookBack  = look_back;
	}

	Py_RETURN_NONE;
    }

    static PyObject *
    Node_get_input_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"input_id", NULL};
        PyObject *input_name_obj = NULL;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "O", kwlist, &input_name_obj)){
            Py_RETURN_NONE;
        }
        const char *input_name = NULL;
        if(input_name_obj){
            input_name = PyStr_AsUTF8(input_name_obj);
        }

	HARK::BufferedNode *bufferedNode = dynamic_cast<HARK::BufferedNode *>(self->node);
	int look_ahead = 0, look_back = 0;
	if(bufferedNode){
	    int input_id = bufferedNode->getInputID(input_name);
	    //std::cout << "Node.cc: get_input_look_window: name=" << bufferedNode->getName() << ", inputName=" << input_name << ", input_id=" << input_id << ", lookWindow=(" << bufferedNode->inputsCache[input_id].lookAhead << ", " << bufferedNode->inputsCache[input_id].lookBack << ")" << std::endl;
	    look_ahead = bufferedNode->inputsCache[input_id].lookAhead;
	    look_back  = bufferedNode->inputsCache[input_id].lookBack;
	}

	return Py_BuildValue("(ii)", look_ahead, look_back);
    }

    static PyObject *
    Node_set_input_look_window(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"input_id", "lookAhead", "lookBack", NULL};
        PyObject *input_name_obj = NULL;
        int look_ahead, look_back;
	if(!PyArg_ParseTupleAndKeywords(args, kwds, "Oii", kwlist, &input_name_obj, &look_ahead, &look_back)){
            Py_RETURN_NONE;
        }
        const char *input_name = NULL;
        if(input_name_obj){
            input_name = PyStr_AsUTF8(input_name_obj);
        }

	HARK::BufferedNode *bufferedNode = dynamic_cast<HARK::BufferedNode *>(self->node);
	if(bufferedNode){
	    int input_id = bufferedNode->getInputID(input_name);
	    //std::cout << "Node.cc: set_input_look_window: name=" << bufferedNode->getName() << ", inputName=" << input_name << ", input_id=" << input_id << ", lookWindow=(" << bufferedNode->inputsCache[input_id].lookAhead << ", " << bufferedNode->inputsCache[input_id].lookBack << ")" << std::endl;
	    bufferedNode->inputsCache[input_id].lookAhead = look_ahead;
	    bufferedNode->inputsCache[input_id].lookBack  = look_back;
	}

	Py_RETURN_NONE;
    }

    static PyObject *
    Node_dump_input(PyObject *_self, PyObject *args, PyObject *kwds)
    {
        Node *self = reinterpret_cast<Node *>(_self);

        /* Parse args */
        static char *kwlist[] = {"count", NULL};
        int count;
        if(!PyArg_ParseTupleAndKeywords(args, kwds, "i", kwlist, &count)){
            Py_RETURN_NONE;
        }

	try{
	    for(int i = 0; i < self->node->inputs.size(); i++){
		HARK::ObjectRef o = self->node->getInput(i, count);
		o->printOn();
		std::cout << std::endl;
	    }

	}HARKMW_CATCH_EXCEPTION();

        Py_RETURN_NONE;
    }

    static PyMemberDef Node_members[] = {
        {NULL}  /* Sentinel */
    };

    static PyMethodDef Node_methods[] = {
        {"get_input_count", (PyCFunction)Node_get_input_count, METH_NOARGS, "Get num of input"},
        {"get_output_count", (PyCFunction)Node_get_output_count, METH_NOARGS, "Get num of output"},
        {"get_output", (PyCFunction)Node_get_output, METH_KEYWORDS | METH_VARARGS, "Get output data from output name and count"},
	{"get_input_look_window", (PyCFunction)Node_get_input_look_window, METH_KEYWORDS | METH_VARARGS, "Get lookAhead/lookBack of output"},
	{"set_input_look_window", (PyCFunction)Node_set_input_look_window, METH_KEYWORDS | METH_VARARGS, "Get lookAhead/lookBack of output"},
	{"get_output_look_window", (PyCFunction)Node_get_output_look_window, METH_KEYWORDS | METH_VARARGS, "Get lookAhead/lookBack of output"},
	{"set_output_look_window", (PyCFunction)Node_set_output_look_window, METH_KEYWORDS | METH_VARARGS, "Get lookAhead/lookBack of output"},
	{"dump_input", (PyCFunction)Node_dump_input, METH_KEYWORDS | METH_VARARGS, "Dump input value"},
        {"connect", (PyCFunction)Node_connect, METH_KEYWORDS | METH_VARARGS, "Connect terminal with name [in] of this node and teminal with name [out] of node [inputNode]"},
        {"process", (PyCFunction)Node_process, METH_KEYWORDS | METH_VARARGS, "Process an output of given count"},
        {"initialize", (PyCFunction)Node_initialize, METH_KEYWORDS | METH_VARARGS, "Initialize node"},
        {NULL}  /* Sentinel */
    };

    PyTypeObject NodeType = {
        PyVarObject_HEAD_INIT(NULL, 0)
        "harkmwnative.Node",             /*tp_name*/
        sizeof(HARKMW::Node),       /*tp_basicsize*/
        0,                         /*tp_itemsize*/
        (destructor)Node_dealloc, /*tp_dealloc*/
        0,                         /*tp_print*/
        0,                         /*tp_getattr*/
        0,                         /*tp_setattr*/
        0,                         /*tp_compare*/
        0,                         /*tp_repr*/
        0,                         /*tp_as_number*/
        0,                         /*tp_as_sequence*/
        0,                         /*tp_as_mapping*/
        0,                         /*tp_hash */
        0,                         /*tp_call*/
        0,                         /*tp_str*/
        0,                         /*tp_getattro*/
        0,                         /*tp_setattro*/
        0,                         /*tp_as_buffer*/
        Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
        "Node objects",            /* tp_doc */
        0,                               /* tp_traverse */
        0,                               /* tp_clear */
        0,                               /* tp_richcompare */
        0,                               /* tp_weaklistoffset */
        0,                               /* tp_iter */
        0,                               /* tp_iternext */
        Node_methods,              /* tp_methods */
        Node_members,              /* tp_members */
        0,                         /* tp_getset */
        0,                         /* tp_base */
        0,                         /* tp_dict */
        0,                         /* tp_descr_get */
        0,                         /* tp_descr_set */
        0,                         /* tp_dictoffset */
        (initproc)Node_init,       /* tp_init */
        0,                         /* tp_alloc */
        Node_new,                  /* tp_new */
    };
}

/* end of file */
