#! /usr/bin/python
# -*- coding: utf-8 -*-

# HARK Middleware
# Copyright (C) 2013-2019 by Honda Research Institute Japan Co., Ltd.
# $Id:$

from harkmw.defs.parameter import Parameters

import harkmw.logic.node
import harkmwnative
from collections import defaultdict


class NativeNode(harkmw.logic.node.Node):
    def __init__(self):
        harkmw.logic.node.Node.__init__(self)
        self.network = None
        self.count = -1
        self.input_count = -1
        self.output_count = -1
        self.input_look_windows = defaultdict(lambda: (0, 0))
        self.output_look_windows = defaultdict(lambda: (0, 0))

    def build(self, network, node_def, params, force_local):
        self.network = network
        self.type = node_def.type
        self.name = node_def.name
        self.topic = node_def.topic
        self.process_name = 'local' if force_local else node_def.get_process_name()

        # Load native module
        self.native_node = harkmwnative.Node(
            self, node_def.type, node_def.parameters)

    def build_type(self, network, type, force_local, name=None):
        # Load native module
        self.native_node = harkmwnative.Node(self, type, Parameters())

        self.network = network
        self.type = type
        self.name = name
        self.topic = None
        self.process_name = 'local' if force_local else network.process_name

    def initialize(self, process):
        self.native_node.initialize()
        harkmw.logic.node.Node.initialize(self, process)

    def release(self):
        self.native_node = None
        self.inputs = {}
        self.outputs = {}

    def process(self, output, count, input_count=-1, output_count=-1):
        self.native_node.process(
            output, count, input_count=input_count, output_count=output_count)
        self.count = count

    def set_output_look_window(self, output_id, look_window):
        self.output_look_windows[output_id] = look_window
        return self.native_node.set_output_look_window(
            output_id, look_window[0], look_window[1])

    def get_output_look_window(self, output_id):
        look_window = self.native_node.get_output_look_window(output_id)
        return max(look_window, self.output_look_windows[output_id])

    def set_input_look_window(self, input_id, look_window):
        self.input_look_windows[input_id] = look_window
        return self.native_node.set_input_look_window(
            input_id, look_window[0], look_window[1])

    def get_input_look_window(self, input_id):
        look_window = self.native_node.get_input_look_window(input_id)
        return max(look_window, self.input_look_windows[input_id])

    def get_input_count(self):
        return self.native_node.get_input_count()

    def get_output_count(self):
        return self.native_node.get_output_count()

    def get_output(self, name, count):
        return self.native_node.get_output(name, count)

    def dump_input(self, count):
        return self.native_node.dump_input(count)

    def connect_node(self, terminal_name, node_from, terminal_from_name):
        self.native_node.connect(
            terminal_name, node_from.native_node, terminal_from_name)

    def dump(self):
        print('[NativeNode]')
        print('  type: %s' % self.type)
        print('  name: %s' % self.name)
        print('  network: type=%s, name=%s' %
              (self.network.type, self.network.name))

        print('  inputs: ')
        for (name, terminal) in self.inputs.items():
            print(
                '    Input: name=%s, node=%s/%s, terminal=%s' %
                (name,
                 terminal.input_node.type,
                 terminal.input_node.name,
                 terminal.input_terminal_name))

        print('  outputs: ')
        for (name, proxy_list) in self.outputs.items():
            for proxy in proxy_list:
                print(
                    '    Output: name=%s, node=%s/%s, terminal=%s' %
                    (name,
                     proxy.output_node.type,
                     proxy.output_node.name,
                     proxy.output_terminal_name))

    def __str__(self):
        return "<%s: type=%s, name=%s>" % (
            self.__class__.__name__, self.type, self.name)

# end of file
